#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
import time
import datetime
import watchdog.events
import watchdog.observers
from ..command import Command
from .. import js, plugins


class JsCommand(Command):

    _arguments = [
        (
            ['mode'],
            {
                'metavar': 'mode',
                'help': 'the type of action to perform',
                'sub-commands':
                    {
                        'shift': (
                            {
                                'help': 'keen to use shifter?'
                            },
                            [
                                (
                                    ['-p', '--plugin'],
                                    {
                                        'action': 'store',
                                        'dest': 'plugin',
                                        'default': None,
                                        'help': 'the name of the plugin or subsystem to target. If not passed, we do our best to guess from the current path.'
                                    }
                                ),
                                (
                                    ['-m', '--module'],
                                    {
                                        'action': 'store',
                                        'dest': 'module',
                                        'default': None,
                                        'help': 'the name of the module in the plugin or subsystem. If omitted all the modules will be shifted, except we are in a module.'
                                    }
                                ),
                                (
                                    ['-w', '--watch'],
                                    {
                                        'action': 'store_true',
                                        'dest': 'watch',
                                        'help': 'watch for changes to re-shift'
                                    }
                                ),
                                (
                                    ['names'],
                                    {
                                        'default': None,
                                        'help': 'name of the instances',
                                        'metavar': 'names',
                                        'nargs': '*'
                                    }
                                )
                            ]
                        ),
                        'doc': (
                            {
                                'help': 'keen to generate documentation?'
                            },
                            [
                                (
                                    ['names'],
                                    {
                                        'default': None,
                                        'help': 'name of the instances',
                                        'metavar': 'names',
                                        'nargs': '*'
                                    }
                                )
                            ]
                        )
                    }
            }
        )
    ]
    _description = 'Wrapper for JS functions'

    def run(self, args):
        if args.mode == 'shift':
            self.shift(args)
        elif args.mode == 'doc':
            self.document(args)


    def shift(self, args):
        """The shift mode"""

        Mlist = self.Wp.resolveMultiple(args.names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        cwd = os.path.realpath(os.path.abspath(os.getcwd()))
        mpath = Mlist[0].get('path')
        relpath = cwd.replace(mpath, '').strip('/')

        # TODO Put that logic somewhere else because it is going to be re-used, I'm sure.
        if not args.plugin:
            (subsystemOrPlugin, pluginName) = plugins.PluginManager.getSubsystemOrPluginFromPath(cwd, Mlist[0])
            if subsystemOrPlugin:
                args.plugin = subsystemOrPlugin + ('_' + pluginName) if pluginName else ''
                logging.info("I guessed the plugin/subsystem to work on as '%s'" % (args.plugin))
            else:
                self.argumentError('The argument --plugin is required, I could not guess it.')

        if not args.module:
            candidate = relpath
            module = None
            while '/yui/src' in candidate:
                (head, tail) = os.path.split(candidate)
                if head.endswith('/yui/src'):
                    module = tail
                    break
                candidate = head

            if module:
                args.module = module
                logging.info("I guessed the JS module to work on as '%s'" % (args.module))

        for M in Mlist:
            if len(Mlist) > 1:
                logging.info('Let\'s shift everything you wanted on \'%s\'' % (M.get('identifier')))

            processor = js.Js(M)
            processor.shift(subsystemOrPlugin=args.plugin, module=args.module)

        if args.watch:
            observer = watchdog.observers.Observer()

            for M in Mlist:
                processor = js.Js(M)
                processorArgs = {'subsystemOrPlugin': args.plugin, 'module': args.module}
                handler = JsShiftWatcher(M, processor, processorArgs)
                observer.schedule(handler, processor.getYUISrcPath(**processorArgs), recursive=True)
                logging.info('Watchdog set up on %s, waiting for changes...' % (M.get('identifier')))

            observer.start()

            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                observer.stop()
            finally:
                observer.join()

    def document(self, args):
        """The docmentation mode"""

        Mlist = self.Wp.resolveMultiple(args.names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        for M in Mlist:
            logging.info('Documenting everything you wanted on \'%s\'. This may take a while...', M.get('identifier'))
            outdir = self.Wp.getExtraDir(M.get('identifier'), 'jsdoc')
            outurl = self.Wp.getUrl(M.get('identifier'), extra='jsdoc')
            processor = js.Js(M)
            processor.document(outdir)
            logging.info('Documentation available at:\n %s\n %s', outdir, outurl)


class JsShiftWatcher(watchdog.events.FileSystemEventHandler):

    _processor = None
    _args = None
    _ext = ['.js', '.json']
    _M = None

    def __init__(self, M, processor, args):
        super(self.__class__, self).__init__()
        self._M = M
        self._processor = processor
        self._args = args

    def on_modified(self, event):
        if event.is_directory:
            return
        elif not os.path.splitext(event.src_path)[1] in self._ext:
            return
        self.process(event)

    def on_moved(self, event):
        if not os.path.splitext(event.dest_path)[1] in self._ext:
            return
        self.process(event)

    def process(self, event):
        logging.info('[%s] (%s) Changes detected!' % (self._M.get('identifier'), datetime.datetime.now().strftime('%H:%M:%S')))

        try:
            self._processor.shift(**self._args)
        except js.ShifterCompileFailed:
            logging.error(' /!\ Error: Compile failed!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

def main():

    import sys
    import argparse
    import os
    import re
    import logging
    import base64
    from .command import CommandRunner
    from .commands import getCommand, commandsList
    from .config import Conf
    from .tools import process
    from .version import __version__

    C = Conf()

    try:
        debuglevel = getattr(logging, C.get('debug').upper())
    except AttributeError:
        debuglevel = logging.INFO

    # Set logging levels.
    logging.basicConfig(format='%(message)s', level=debuglevel)
    logging.getLogger('requests').setLevel(logging.WARNING)  # Reset logging level of 'requests' module.
    logging.getLogger('keyring.backend').setLevel(logging.WARNING)

    availaliases = [str(x) for x in list(C.get('aliases').keys())]
    choices = sorted(commandsList + availaliases)

    parser = argparse.ArgumentParser(description='Moodle Development Kit', add_help=False)
    parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
    parser.add_argument('-l', '--list', action='store_true', help='list the available commands')
    parser.add_argument('-v', '--version', action='store_true', help='display the current version')
    parser.add_argument(*['--%s'%base64.b64decode(f) for f in ('aWNhbnRyZWFj', 'aWNhbnRyZWFk')], dest='asdf', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('command', metavar='command', nargs='?', help='command to call', choices=choices)
    parser.add_argument('args', metavar='arguments', nargs=argparse.REMAINDER, help='arguments of the command')
    parsedargs = parser.parse_args()

    cmd = parsedargs.command
    args = parsedargs.args

    # There is no command, what do we do?
    if not cmd:
        if parsedargs.version:
            print('MDK version %s' % __version__)
        elif parsedargs.asdf:
            print(base64.b64decode('U29ycnkgRGF2ZSwgTURLIGNhbm5vdCBoZWxwIHlvdSB3aXRoIHRoYXQuLi4='))
        elif parsedargs.list:
            for c in sorted(commandsList):
                print('{0:<15} {1}'.format(c, getCommand(c)._description))
        else:
            parser.print_help()
        sys.exit(0)

    # Looking up for an alias
    alias = C.get('aliases.%s' % cmd)
    if alias != None:
        if alias.startswith('!'):
            cmd = alias[1:]
            i = 0
            # Replace $1, $2, ... with passed arguments
            for arg in args:
                i += 1
                cmd = cmd.replace('$%d' % i, arg)
            # Remove unknown $[0-9]
            cmd = re.sub(r'\$[0-9]', '', cmd)
            result = process(cmd, stdout=None, stderr=None)
            sys.exit(result[0])
        else:
            cmd = alias.split(' ')[0]
            args = alias.split(' ')[1:] + args

    cls = getCommand(cmd)
    Cmd = cls(C)
    Runner = CommandRunner(Cmd)
    try:
        Runner.run(args, prog='%s %s' % (os.path.basename(sys.argv[0]), cmd))
    except Exception as e:
        import traceback
        info = sys.exc_info()
        logging.error('%s: %s', e.__class__.__name__, e)
        logging.debug(''.join(traceback.format_tb(info[2])))
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..command import Command
from ..scripts import Scripts


class RunCommand(Command):

    _arguments = [
        (
            ['-l', '--list'],
            {
                'action': 'store_true',
                'dest': 'list',
                'help': 'list the available scripts'
            }
        ),
        (
            ['-a', '--all'],
            {
                'action': 'store_true',
                'dest': 'all',
                'help': 'runs the script on each instance'
            }
        ),
        (
            ['-i', '--integration'],
            {
                'action': 'store_true',
                'dest': 'integration',
                'help': 'runs the script on integration instances'
            }
        ),
        (
            ['-s', '--stable'],
            {
                'action': 'store_true',
                'dest': 'stable',
                'help': 'runs the script on stable instances'
            }
        ),
        (
            ['-g', '--arguments'],
            {
                'help': 'a list of arguments to pass to the script. Use --arguments="--list of --arguments" if you need to use dashes. Otherwise add -- after the argument list.',
                'metavar': 'arguments',
                'nargs': '+'
            }
        ),
        (
            ['script'],
            {
                'nargs': '?',
                'help': 'the name of the script to run'
            }
        ),
        (
            ['names'], {
                'default': None,
                'help': 'name of the instances',
                'nargs': '*'
            }
        )
    ]
    _description = 'Run a script on a Moodle instance'

    def run(self, args):

        # Printing existing scripts
        if args.list:
            scripts = Scripts.list()
            for script in sorted(scripts.keys()):
                print('%s (%s)' % (script, scripts[script]))
            return

        # Trigger error when script is missing
        if not args.script:
            self.argumentError('missing script name')

        # Resolving instances
        names = args.names
        if args.all:
            names = self.Wp.list()
        elif args.integration or args.stable:
            names = self.Wp.list(integration=args.integration, stable=args.stable)

        # Doing stuff
        Mlist = self.Wp.resolveMultiple(names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        for M in Mlist:
            logging.info('Running \'%s\' on \'%s\'' % (args.script, M.get('identifier')))
            try:
                M.runScript(args.script, stderr=None, stdout=None, arguments=args.arguments)
            except Exception as e:
                logging.warning('Error while running the script on %s' % M.get('identifier'))
                logging.debug(e)
            else:
                logging.info('')

        logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
from .tools import process
from .config import Conf

C = Conf()


class Css(object):
    """Class wrapping CSS related functions"""

    _M = None

    _debug = False
    _compiler = 'grunt'

    def __init__(self, M):
        self._M = M

    def setCompiler(self, compiler):
        self._compiler = compiler

    def setDebug(self, debug):
        self._debug = debug

    def compile(self, theme='bootstrapbase', sheets=None):
        """Compile LESS sheets contained within a theme"""

        source = self.getThemeLessPath(theme)
        dest = self.getThemeCssPath(theme)
        if not os.path.isdir(source):
            raise Exception('Unknown theme %s, or less directory not found' % (theme))

        if not sheets:
            # Guess the sheets from the theme less folder.
            sheets = []
            for candidate in os.listdir(source):
                if os.path.isfile(os.path.join(source, candidate)) and candidate.endswith('.less'):
                    sheets.append(os.path.splitext(candidate)[0])
        elif type(sheets) != list:
            sheets = [sheets]

        if len(sheets) < 1:
            logging.warning('Could not find any sheets')
            return False

        hadErrors = False

        if self._compiler == 'grunt':
            sheets = ['moodle']

        for name in sheets:
            sheet = name + '.less'
            destSheet = name + '.css'

            if not os.path.isfile(os.path.join(source, sheet)):
                logging.warning('Could not find file %s' % (sheet))
                hadErrors = True
                continue

            try:
                if self._compiler == 'grunt':
                    compiler = Grunt(source, os.path.join(source, sheet), os.path.join(dest, destSheet))
                elif self._compiler == 'recess':
                    compiler = Recess(source, os.path.join(source, sheet), os.path.join(dest, destSheet))
                elif self._compiler == 'lessc':
                    compiler = Lessc(self.getThemeDir(), os.path.join(source, sheet), os.path.join(dest, destSheet))

                compiler.setDebug(self._debug)

                compiler.execute()
            except CssCompileFailed:
                logging.warning('Failed compilation of %s' % (sheet))
                hadErrors = True
                continue
            else:
                logging.info('Compiled %s to %s' % (sheet, destSheet))

        return not hadErrors

    def getThemeCssPath(self, theme):
        return os.path.join(self.getThemePath(theme), 'style')

    def getThemeLessPath(self, theme):
        return os.path.join(self.getThemePath(theme), 'less')

    def getThemeDir(self):
        return os.path.join(self._M.get('path'), 'theme')

    def getThemePath(self, theme):
        return os.path.join(self.getThemeDir(), theme)


class Compiler(object):
    """LESS compiler abstract"""

    _compress = True
    _debug = False
    _cwd = None
    _source = None
    _dest = None

    def __init__(self, cwd, source, dest):
        self._cwd = cwd
        self._source = source
        self._dest = dest

    def execute(self):
        raise Exception('Compiler does not implement execute() method')

    def setCompress(self, compress):
        self._compress = compress

    def setDebug(self, debug):
        self._debug = debug


class Grunt(Compiler):
    """Grunt compiler"""

    def execute(self):
        executable = C.get('grunt')
        if not executable:
            raise Exception('Could not find executable path')

        cmd = [executable, 'css']

        (code, out, err) = process(cmd, self._cwd)
        if code != 0 or len(out) == 0:
            raise CssCompileFailed('Error during compile')


class Recess(Compiler):
    """Recess compiler"""

    def execute(self):
        executable = C.get('recess')
        if not executable:
            raise Exception('Could not find executable path')

        cmd = [executable, self._source, '--compile']

        if self._compress:
            cmd.append('--compress')

        (code, out, err) = process(cmd, self._cwd)
        if code != 0 or len(out) == 0:
            raise CssCompileFailed('Error during compile')

        # Saving to destination
        with open(self._dest, 'w') as f:
            f.write(out)


class Lessc(Compiler):
    """Lessc compiler"""

    def execute(self):
        executable = C.get('lessc')
        if not executable:
            raise Exception('Could not find executable path')

        cmd = [executable]

        sourcePath = os.path.relpath(self._source, self._cwd)
        sourceDir = os.path.dirname(sourcePath)

        if self._debug:
            cmd.append('--source-map-rootpath=' + sourceDir)
            cmd.append('--source-map-map-inline')
            self.setCompress(False)

        if self._compress:
            cmd.append('--compress')

        # Append the source and destination.
        cmd.append(sourcePath)
        cmd.append(os.path.relpath(self._dest, self._cwd))

        (code, out, err) = process(cmd, self._cwd)
        if code != 0 or len(out) != 0:
            raise CssCompileFailed('Error during compile')


class CssCompileFailed(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..command import Command


class PurgeCommand(Command):

    _arguments = [
        (
            ['-a', '--all'],
            {
                'action': 'store_true',
                'dest': 'all',
                'help': 'purge the cache on each instance'
            }
        ),
        (
            ['-i', '--integration'],
            {
                'action': 'store_true',
                'dest': 'integration',
                'help': 'purge the cache on integration instances'
            }
        ),
        (
            ['-s', '--stable'],
            {
                'action': 'store_true',
                'dest': 'stable',
                'help': 'purge the cache on stable instances'
            }
        ),
        (
            ['-m', '--manual'],
            {
                'action': 'store_true',
                'dest': 'manual',
                'help': 'perform a manual deletion of some cache in dataroot before executing the CLI script'
            }
        ),
        (
            ['names'],
            {
                'default': None,
                'help': 'name of the instances',
                'metavar': 'names',
                'nargs': '*'
            }
        )
    ]
    _description = 'Purge the cache of an instance'

    def run(self, args):

        # Resolving instances
        names = args.names
        if args.all:
            names = self.Wp.list()
        elif args.integration or args.stable:
            names = self.Wp.list(integration=args.integration, stable=args.stable)

        # Doing stuff
        Mlist = self.Wp.resolveMultiple(names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        for M in Mlist:
            logging.info('Purging cache on %s' % (M.get('identifier')))

            try:
                M.purge(manual=args.manual)
            except Exception as e:
                logging.error('Could not purge cache: %s' % e)
            else:
                logging.debug('Cache purged!')

            logging.info('')

        logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import shutil
import logging
from .tools import mkdir, process, stableBranch
from .exceptions import CreateException
from .config import Conf
from . import git
from . import moodle

C = Conf()


class Workplace(object):

    """The name of the directory that contains the PHP files"""
    wwwDir = None

    """The name of the directory that contains Moodle data"""
    dataDir = None

    """The name of the directory that contains extra files"""
    extraDir = None

    """The name of the directory that makes extraDir web accessible, see getMdkWebDir"""
    mdkDir = None

    """The path to the storage directory"""
    path = None

    """The path to MDK cache"""
    cache = None

    """The path to the web accessible directory"""
    www = None

    def __init__(self, path=None, wwwDir=None, dataDir=None, extraDir=None, mdkDir=None):
        if path == None:
            path = C.get('dirs.storage')
        if wwwDir == None:
            wwwDir = C.get('wwwDir')
        if dataDir == None:
            dataDir = C.get('dataDir')
        if extraDir == None:
            extraDir = C.get('extraDir')
        if mdkDir == None:
            mdkDir = C.get('mdkDir')

        # Directory paths
        self.path = os.path.abspath(os.path.realpath(os.path.expanduser(path)))
        self.cache = os.path.abspath(os.path.realpath(os.path.expanduser(C.get('dirs.mdk'))))
        self.www = os.path.abspath(os.path.realpath(os.path.expanduser(C.get('dirs.www'))))

        if not os.path.isdir(self.path):
            raise Exception('Directory %s not found' % self.path)

        # Directory names
        self.wwwDir = wwwDir
        self.dataDir = dataDir
        self.extraDir = extraDir
        self.mdkDir = mdkDir

    def checkCachedClones(self, stable=True, integration=True):
        """Clone the official repository in a local cache"""
        cacheStable = self.getCachedRemote(False)
        cacheIntegration = self.getCachedRemote(True)

        if not os.path.isdir(cacheStable) and stable:
            logging.info('Cloning stable repository into cache...')

            # For faster clone, we will copy the integration clone if it exists.
            if os.path.isdir(cacheIntegration):
                shutil.copytree(cacheIntegration, cacheStable)
                repo = git.Git(cacheStable, C.get('git'))
                repo.setRemote('origin', C.get('remotes.stable'))
                # The repository is not updated at this stage, it has to be done manually.
            else:
                logging.info('This is going to take a while...')
                process('%s clone --mirror %s %s' % (C.get('git'), C.get('remotes.stable'), cacheStable))

        if not os.path.isdir(cacheIntegration) and integration:
            logging.info('Cloning integration repository into cache...')

            # For faster clone, we will copy the integration clone if it exists.
            if os.path.isdir(cacheStable):
                shutil.copytree(cacheStable, cacheIntegration)
                repo = git.Git(cacheIntegration, C.get('git'))
                repo.setRemote('origin', C.get('remotes.integration'))
                # The repository is not updated at this stage, it has to be done manually.
            else:
                logging.info('Have a break, this operation is slow...')
                process('%s clone --mirror %s %s' % (C.get('git'), C.get('remotes.integration'), cacheIntegration))

    def create(self, name=None, version='main', integration=False, useCacheAsRemote=False):
        """Creates a new instance of Moodle.
        The parameter useCacheAsRemote has been deprecated.
        """
        if name == None:
            name = self.generateInstanceName(version, integration=integration)

        if name == self.mdkDir:
            raise Exception('A Moodle instance cannot be called \'%s\', this is a reserved word.' % self.mdkDir)

        installDir = self.getPath(name)
        wwwDir = self.getPath(name, 'www')
        dataDir = self.getPath(name, 'data')
        extraDir = self.getPath(name, 'extra')
        linkDir = os.path.join(self.www, name)
        extraLinkDir = os.path.join(self.getMdkWebDir(), name)

        # Check the cached clones and create them if necessary.
        self.checkCachedClones(not integration, integration)
        # Update the cached clones.
        self.updateCachedClones(stable=not integration, integration=integration, verbose=False)

        branch = stableBranch(version, git.Git(self.getCachedRemote(integration), C.get('git')))

        useCacheAsUpstream = C.get('useCacheAsUpstreamRemote')
        cloneAsShared = useCacheAsUpstream and C.get('useCacheAsSharedClone')

        if self.isMoodle(name):
            raise CreateException('The Moodle instance %s already exists' % name)
        elif os.path.isdir(installDir):
            raise CreateException('Installation path exists: %s' % installDir)

        mkdir(installDir, 0o755)
        mkdir(wwwDir, 0o755)
        mkdir(dataDir, 0o777)
        mkdir(extraDir, 0o777)

        repository = self.getCachedRemote(integration)

        # Clone the instances
        logging.info('Cloning repository...')
        process(f'{C.get("git")} clone --branch {branch} --single-branch '
                f'{"--shared" if cloneAsShared else ""} {repository} {wwwDir}')

        # Symbolic link
        if os.path.islink(linkDir):
            os.remove(linkDir)
        if os.path.isfile(linkDir) or os.path.isdir(linkDir):  # No elif!
            logging.warning('Could not create symbolic link. Please manually create: ln -s %s %s' % (wwwDir, linkDir))
        else:
            os.symlink(wwwDir, linkDir)

        # Symlink to extra.
        if os.path.isfile(extraLinkDir) or os.path.isdir(extraLinkDir):
            logging.warning('Could not create symbolic link. Please manually create: ln -s %s %s' % (extraDir, extraLinkDir))
        else:
            os.symlink(extraDir, extraLinkDir)

        # Symlink to dataDir in wwwDir
        if type(C.get('symlinkToData')) == str:
            linkDataDir = os.path.join(wwwDir, C.get('symlinkToData'))
            if not os.path.isfile(linkDataDir) and not os.path.isdir(linkDataDir) and not os.path.islink(linkDataDir):
                os.symlink(dataDir, linkDataDir)

        logging.info('Checking out branch...')
        repo = git.Git(wwwDir, C.get('git'))

        # Removing the default remote origin coming from the clone
        repo.delRemote('origin')

        # Setting up the correct remote names
        repo.setRemote(C.get('myRemote'), C.get('remotes.mine'))
        repo.setRemote(C.get('upstreamRemote'), repository)

        # Creating, fetch, pulling branches
        repo.fetch(C.get('upstreamRemote'))
        track = '%s/%s' % (C.get('upstreamRemote'), branch)
        if not repo.hasBranch(branch) and not repo.createBranch(branch, track):
            logging.error('Could not create branch %s tracking %s' % (branch, track))
        else:
            repo.checkout(branch)
        repo.pull(remote=C.get('upstreamRemote'))

        # Fixing up remote URLs if need be, this is done after pulling the cache one because we
        # do not want to contact the real origin server from here, it is slow and pointless.
        if not useCacheAsUpstream:
            realupstream = C.get('remotes.integration') if integration else C.get('remotes.stable')
            if realupstream:
                repo.setRemote(C.get('upstreamRemote'), realupstream)

        M = self.get(name)
        return M

    def delete(self, name):
        """Completely remove an instance, database included"""

        # Instantiating the object also checks if it exists
        M = self.get(name)

        # Deleting the whole thing
        shutil.rmtree(os.path.join(self.path, name))

        # Deleting the possible symlink
        link = os.path.join(self.www, name)
        if os.path.islink(link):
            try:
                os.remove(link)
            except Exception:
                pass

        # Delete the extra dir symlink
        link = os.path.join(self.getMdkWebDir(), name)
        if os.path.islink(link):
            try:
                os.remove(link)
            except Exception:
                pass

        # Delete db
        DB = M.dbo()
        dbname = M.get('dbname')
        if DB and dbname and DB.dbexists(dbname):
            DB.dropdb(dbname)

    def generateInstanceName(self, version, integration=False, suffix='', identifier=None):
        """Creates a name (identifier) from arguments"""

        if identifier != None:
            # If an identifier is passed, we use it regardless of the other parameters.
            # Except for suffix.
            name = identifier.replace(' ', '_')
        else:
            # Wording version
            prefixmaster = C.get('wording.prefixMaster')
            if version == 'master':
                prefixVersion = prefixmaster
            elif version == 'main':
                prefixmain = C.get('wording.prefixMain')
                # Check if we need to sync master and main prefix wordings. Should be a one-off thing.
                if not C.get('wordingPrefixesChecked'):
                    # If wording.prefixMaster has been customised while wording.prefixMain remains default,
                    # copy wording.prefixMaster to wording.prefixMain to maintain customisation.
                    if prefixmaster != 'master' and prefixmain == 'main':
                        prefixmain = prefixmaster
                        C.set('wording.prefixMain', prefixmain)
                        logging.info('The config wording.prefixMaster (%s) has been copied to wording.prefixMain'
                                     % prefixmain)
                    # Set config wordingPrefixesChecked to true. We shouldn't need to do this next time.
                    C.set('wordingPrefixesChecked', True)
                prefixVersion = prefixmain
            else:
                prefixVersion = version

            # Generating name
            if integration:
                name = C.get('wording.prefixIntegration') + prefixVersion
            else:
                name = C.get('wording.prefixStable') + prefixVersion

        # Append the suffix
        if suffix != None and suffix != '':
            name += C.get('wording.suffixSeparator') + suffix

        return name

    def get(self, name):
        """Returns an instance defined by its name, or by path"""
        # Extracts name from path
        if os.sep in name:
            path = os.path.abspath(os.path.realpath(name))
            if not path.startswith(self.path):
                raise Exception('Could not find Moodle instance at %s' % name)
            (head, name) = os.path.split(path)

        if not self.isMoodle(name):
            raise Exception('Could not find Moodle instance %s' % name)
        return moodle.Moodle(os.path.join(self.path, name, self.wwwDir), identifier=name)

    def getCachedRemote(self, integration=False):
        """Return the path to the cached remote"""
        if integration:
            return os.path.join(self.cache, 'integration.git')
        else:
            return os.path.join(self.cache, 'moodle.git')

    def getExtraDir(self, name, subdir=None):
        """Return the path to the extra directory of an instance

        This also creates the directory if does not exist.
        """
        path = self.getPath(name, 'extra')
        if subdir:
            path = os.path.join(path, subdir)
        if not os.path.exists(path):
            mkdir(path, 0o777)
        return path

    def getMdkWebDir(self):
        """Return (and create) the special MDK web directory."""
        mdkExtra = os.path.join(self.www, self.mdkDir)
        if not os.path.exists(mdkExtra):
            mkdir(mdkExtra, 0o777)

        return mdkExtra

    def getPath(self, name, mode=None):
        """Returns the path of an instance base on its name"""
        base = os.path.join(self.path, name)
        if mode == 'www':
            return os.path.join(base, self.wwwDir)
        elif mode == 'data':
            return os.path.join(base, self.dataDir)
        elif mode == 'extra':
            return os.path.join(base, self.extraDir)
        else:
            return base

    def getUrl(self, name, extra=None):
        """Return the URL to an instance, or to its extra directory if extra is passed"""
        base = '%s://%s' % (C.get('scheme'), C.get('host'))

        if C.get('path') != '' and C.get('path') != None:
            base = '%s/%s' % (base, C.get('path'))

        wwwroot = None
        if not extra:
            wwwroot = '%s/%s' % (base, name)
        else:
            wwwroot = '%s/%s/%s/%s' % (base, self.mdkDir, name, extra)

        return wwwroot

    def isMoodle(self, name):
        """Checks whether a Moodle instance exist under this name"""
        d = os.path.join(self.path, name)
        if not os.path.isdir(d):
            return False

        wwwDir = os.path.join(d, self.wwwDir)
        dataDir = os.path.join(d, self.dataDir)
        if not os.path.isdir(wwwDir) or not os.path.isdir(dataDir):
            return False

        if not moodle.Moodle.isInstance(wwwDir):
            return False

        return True

    def list(self, integration=None, stable=None):
        """Return the list of Moodle instances"""
        dirs = os.listdir(self.path)
        names = []
        for d in dirs:
            if d == '.' or d == '..': continue
            if not os.path.isdir(os.path.join(self.path, d)): continue
            if not self.isMoodle(d): continue
            if integration != None or stable != None:
                M = self.get(d)
                if not integration and M.isIntegration(): continue
                if not stable and M.isStable(): continue
            names.append(d)
        return names

    def resolve(self, name=None, path=None):
        """Try to find a Moodle instance based on its name, a path or the working directory"""

        # A name was passed, is that a valid instance?
        if name != None:
            if self.isMoodle(name):
                return self.get(name)
            return None

        # If a path was not passed, let's use the current working directory.
        if path == None:
            path = os.getcwd()
        path = os.path.realpath(os.path.abspath(path))

        # Is this path in a Moodle instance?
        if path.startswith(self.path):

            # Get the relative path identifier/some/other/path
            relative = os.path.relpath(path, self.path)

            # Isolating the identifier, it should be the first directory
            (head, tail) = os.path.split(relative)
            while head:
                (head, tail) = os.path.split(head)

            if self.isMoodle(tail):
                return self.get(tail)

        return False

    def resolveMultiple(self, names=[]):
        """Return multiple instances"""
        if type(names) != list:
            if type(names) == str:
                names = list(names)
            else:
                raise Exception('Unexpected variable type')

        # Nothing has been passed, we use resolve()
        if len(names) < 1:
            M = self.resolve()
            if M:
                return [M]
            else:
                return []

        # Try to resolve each instance
        result = []
        for name in names:
            M = self.resolve(name=name)
            if M:
                result.append(M)
            else:
                logging.info('Could not find instance called %s' % name)
        return result

    def updateCachedClones(self, integration=True, stable=True, verbose=True):
        """Update the cached clone of the repositories"""

        caches = []

        if integration:
            caches.append(os.path.join(self.cache, 'integration.git'))
        if stable:
            caches.append(os.path.join(self.cache, 'moodle.git'))

        for cache in caches:
            if not os.path.isdir(cache):
                continue

            repo = git.Git(cache, C.get('git'))

            if verbose:
                logging.info('Fetching cached repository %s...', os.path.basename(cache))
            else:
                logging.debug('Fetching cached repository %s...', os.path.basename(cache))
            if not repo.fetch():
                raise Exception('Could not fetch in repository %s' % (cache))

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import json
import time
import logging
from distutils.dir_util import copy_tree

from .tools import chmodRecursive, mkdir
from .db import DB
from .config import Conf
from .workplace import Workplace
from .exceptions import *

C = Conf()
jason = 'info.json'
sqlfile = 'dump.sql'


class BackupManager(object):

    def __init__(self):
        self.path = os.path.expanduser(os.path.join(C.get('dirs.moodle'), 'backup'))
        if not os.path.exists(self.path):
            mkdir(self.path, 0o777)

    def create(self, M):
        """Creates a new backup of M"""

        if M.isInstalled() and M.get('dbtype') not in ('mysqli', 'mariadb'):
            raise BackupDBEngineNotSupported('Cannot backup database engine %s' % M.get('dbtype'))

        name = M.get('identifier')
        if name == None:
            raise Exception('Cannot backup instance without identifier!')

        now = int(time.time())
        backup_identifier = self.createIdentifier(name)
        Wp = Workplace()

        # Copy whole directory, shutil will create topath
        topath = os.path.join(self.path, backup_identifier)
        path = Wp.getPath(name)
        logging.info('Copying instance directory')
        copy_tree(path, topath, preserve_symlinks=1)

        # Dump the whole database
        if M.isInstalled():
            logging.info('Dumping database')
            dumpto = os.path.join(topath, sqlfile)
            fd = open(dumpto, 'w')
            M.dbo().selectdb(M.get('dbname'))
            M.dbo().dump(fd)
        else:
            logging.info('Instance not installed. Do not dump database.')

        # Create a JSON file containing all known information
        logging.info('Saving instance information')
        jsonto = os.path.join(topath, jason)
        info = M.info()
        info['backup_origin'] = path
        info['backup_identifier'] = backup_identifier
        info['backup_time'] = now
        json.dump(info, open(jsonto, 'w'), sort_keys=True, indent=4)

        return True

    def createIdentifier(self, name):
        """Creates an identifier"""
        for i in range(1, 100):
            identifier = '{0}_{1:0>2}'.format(name, i)
            if not self.exists(identifier):
                break
            identifier = None
        if not identifier:
            raise Exception('Could not generate a backup identifier! How many backup did you do?!')
        return identifier

    def exists(self, name):
        """Checks whether a backup exists under this name or not"""
        d = os.path.join(self.path, name)
        f = os.path.join(d, jason)
        if not os.path.isdir(d):
            return False
        return os.path.isfile(f)

    def get(self, name):
        return Backup(self.getPath(name))

    def getPath(self, name):
        return os.path.join(self.path, name)

    def list(self):
        """Returns a list of backups with their information"""
        dirs = os.listdir(self.path)
        backups = {}
        for name in dirs:
            if name == '.' or name == '..': continue
            if not self.exists(name): continue
            try:
                backups[name] = Backup(self.getPath(name))
            except:
                # Must successfully retrieve information to be a valid backup
                continue
        return backups


class Backup(object):

    def __init__(self, path):
        self.path = path
        self.jason = os.path.join(path, jason)
        self.sqlfile = os.path.join(path, sqlfile)
        if not os.path.isdir(path):
            raise Exception('Could not find backup in %s' % path)
        elif not os.path.isfile(self.jason):
            raise Exception('Backup information file unfound!')
        self.load()

    def get(self, name):
        """Returns a info on the backup"""
        try:
            return self.infos[name]
        except:
            return None

    def load(self):
        """Loads the backup information"""
        if not os.path.isfile(self.jason):
            raise Exception('Backup information file not found!')
        try:
            self.infos = json.load(open(self.jason, 'r'))
        except:
            raise Exception('Could not load information from JSON file')

    def restore(self, destination=None):
        """Restores the backup"""

        identifier = self.get('identifier')
        if not identifier:
            raise Exception('Identifier is invalid! Cannot proceed.')

        Wp = Workplace()
        if destination == None:
            destination = self.get('backup_origin')
        if not destination:
            raise Exception('Wrong path to perform the restore!')

        if os.path.isdir(destination):
            raise BackupDirectoryExistsException('Destination directory already exists!')

        # Restoring database
        if self.get('installed') and os.path.isfile(self.sqlfile):
            dbname = self.get('dbname')
            dbo = DB(self.get('dbtype'), C.get('db.%s' % self.get('dbtype')))
            if dbo.dbexists(dbname):
                raise BackupDBExistsException('Database already exists!')

        # Copy tree to destination
        try:
            logging.info('Restoring instance directory')
            copy_tree(self.path, destination, preserve_symlinks=1)
            M = Wp.get(identifier)
            chmodRecursive(Wp.getPath(identifier, 'data'), 0o777)
        except Exception as e:
            raise Exception('Error while restoring directory\n%s\nto %s. Exception: %s' % (self.path, destination, e))

        # Restoring database
        if self.get('installed') and os.path.isfile(self.sqlfile):
            logging.info('Restoring database')
            content = ''
            f = open(self.sqlfile, 'r')
            for l in f:
                content += l
            queries = content.split(';\n')
            content = None
            logging.info("%d queries to execute" % (len(queries)))

            dbo.createdb(dbname)
            dbo.selectdb(dbname)
            done = 0
            for query in queries:
                if len(query.strip()) == 0: continue
                try:
                    dbo.execute(query)
                except:
                    logging.error('Query failed! You will have to fix this mually. %s', query)
                done += 1
                if done % 500 == 0:
                    logging.debug("%d queries done" % done)
            logging.info('%d queries done' % done)
            dbo.close()

        # Restoring symbolic link
        linkDir = os.path.join(Wp.www, identifier)
        wwwDir = Wp.getPath(identifier, 'www')
        if os.path.islink(linkDir):
            os.remove(linkDir)
        if os.path.isfile(linkDir) or os.path.isdir(linkDir):  # No elif!
            logging.warning('Could not create symbolic link. Please manually create: ln -s %s %s' % (wwwDir, linkDir))
        else:
            os.symlink(wwwDir, linkDir)

        return M



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
from setuptools import setup, find_packages

# Load version number.
from mdk.version import __version__

# Get the long description from the relevant file.
longDescription = ''
with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    longDescription = f.read()

# Load the requirements.
requirements = []
with open('requirements.txt') as f:
    requirements = f.readlines()

# Get the content of the scripts folder.
scripts = []
for f in os.listdir(os.path.join(os.path.dirname(__file__), 'mdk', 'scripts')):
    if f == 'README.rst':
        continue
    scripts.append('scripts/%s' % (f))

setup(
    name='moodle-sdk',
    version=__version__,
    description='Moodle Development Kit',
    long_description=longDescription,
    license='MIT',

    url='https://github.com/FMCorz/mdk',
    author='Frédéric Massart',
    author_email='fred@fmcorz.net',
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Education',
        'Topic :: Software Development',
        'Topic :: Utilities'
    ],
    keywords='mdk moodle moodle-sdk',

    packages=find_packages(),
    package_data={'mdk': ['config-dist.json'] + scripts},
    install_requires=requirements,
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'mdk = mdk.__main__:main'
        ]
    }
)
